#include "vaultscript.h"
#include <map>
#include <cstdio>
#include <cstring>
#include <cmath>
#include <ctime>
#include <cstdlib>

using namespace vaultmp;

enum class AntState : unsigned int
{
	ANT_IDLE = 0x01,
	ANT_ALERT = ANT_IDLE << 1,
	ANT_AGGRO = ANT_ALERT << 1,
	ANT_DEAD = ANT_AGGRO << 1,
};

static std::vector<IDLE> antsReactions {
	IDLE::AntHitHead,
	IDLE::AntDownFaceRight,
	IDLE::AntDownFaceLeft,
	IDLE::AntBrain,
	IDLE::AntHitRightLeg,
	IDLE::AntHitLeftLegs,
};

static Base antType = static_cast<Base>(CREA::CrAnt2Soldier);
static IDHash<AntState> antsAI;
static IDHash<Value> antsHealth;
static IDHash<ID> antsTarget;
static ActorValue health = ActorValue::Health;
static Value antHealth = 150.0;
static Value antAlert = 500.0;
static Value antAggro = 700.0;
static Value antDamage = 0.3;
static Value antSpeed = 30.0;

Void VAULTSCRIPT exec() noexcept
{
	std::srand(std::time(nullptr));
}

void AdvancePosition(Value& X, Value& Y, Value aZ, Value offset)
{
	X += (offset * std::sin((aZ / 360.0) * 2 * M_PI));
	Y += (offset * std::cos((aZ / 360.0) * 2 * M_PI));
}

Result RemoveObject(ID object_) noexcept
{
	DestroyObject(object_);
	antsAI.erase(object_);
	antsTarget.erase(object_);
	antsHealth.erase(object_);
	KillTimer();
	return static_cast<Result>(0);
}

Result AntAI(ID ant_) noexcept
{
	Actor ant(ant_);

	if (ant)
	{
		if (ant.GetActorDead())
		{
			antsAI[ant_] = AntState::ANT_DEAD;
			Chat << "The ant has been killed";
			KillTimer();
			CreateTimerEx(RemoveObject, static_cast<Interval>(8000), ant_);
		}
		else
		{
			if (antsAI[ant_] == AntState::ANT_AGGRO)
			{
				Actor actor(antsTarget[ant_]);

				if (actor && !actor.GetActorDead())
				{
					Value actorX, actorY, actorZ;
					Value antX, antY, antZ, antAX, antAY, antAZ;

					actor.GetPos(actorX, actorY, actorZ);
					ant.GetPos(antX, antY, antZ);
					ant.GetAngle(antAX, antAY, antAZ);

					antAZ = atan2(actorX - antX, actorY - antY) * 180 / M_PI;

					if (antAZ < 0)
						antAZ += 360.0;

					ant.SetAngle(antAX, antAY, antAZ);

					if (!ant.IsNearPoint(actorX, actorY, actorZ, 80.0))
					{
						AdvancePosition(antX, antY, antAZ, antSpeed);
						ant.SetPos(antX, antY, antZ);
					}
					else
					{
						Value hp = actor.GetActorValue(health);

						if (hp <= antDamage)
						{
							Chat << "The ant killed " + actor.GetBaseName();
							actor.KillActor();
						}
						else
							actor.SetActorValue(health, hp - antDamage);
					}
				}
				else
					antsAI[ant_] = AntState::ANT_IDLE;
			}
			else
			{
				auto actors = Actor::GetList();
				auto players = Player::GetList();

				actors.insert(actors.end(), players.begin(), players.end());

				for (const auto& actor_ : actors)
				{
					Actor actor(actor_);

					if (actor.GetBase() == antType)
						continue;

					Value X, Y, Z;
					actor.GetPos(X, Y, Z);

					if (actor.GetCell() == ant.GetCell() && ant.IsNearPoint(X, Y, Z, antAlert) && antsAI[ant_] == AntState::ANT_IDLE)
					{
						Chat << String(actor.GetBaseName() + " raised the attention of the ant!");
						antsAI[ant_] = AntState::ANT_ALERT;
						ant.PlayIdle(IDLE::Antenna);
					}
				}
			}
		}
	}
	else
		KillTimer();

	return static_cast<Result>(0);
}

State VAULTSCRIPT OnPlayerChat(ID player, RawString message) noexcept
{
	if (!std::strcmp(message, "/ant"))
	{
		static std::map<unsigned char, std::pair<double, double>> values {
			{0x00, {2.000000, 2.000000}},
			{0x01, {4.000000, 4.000000}},
			{0x02, {50.000000, 50.000000}},
			{0x03, {50.000000, 50.000000}},
			{0x05, {7.000000, 7.000000}},
			{0x06, {4.000000, 4.000000}},
			{0x07, {6.000000, 6.000000}},
			{0x08, {2.000000, 2.000000}},
			{0x09, {2.000000, 2.000000}},
			{0x0A, {6.000000, 6.000000}},
			{0x0B, {5.000000, 5.000000}},
			{0x0D, {220.000000, 220.000000}},
			{0x0E, {5.000000, 5.000000}},
			{0x0F, {5.000000, 5.000000}},
			{0x10, {antHealth, antHealth}},
			{0x11, {3.000000, 3.000000}},
			{0x13, {25.000000, 25.000000}},
			{0x14, {10.000000, 10.000000}},
			{0x15, {100.000000, 100.000000}},
			{0x16, {100.000000, 100.000000}},
			{0x17, {-500.000000, -500.000000}},
			{0x19, {100.000000, 100.000000}},
			{0x1A, {100.000000, 100.000000}},
			{0x1B, {100.000000, 100.000000}},
			{0x1C, {100.000000, 100.000000}},
			{0x1D, {100.000000, 100.000000}},
			{0x1E, {100.000000, 100.000000}},
			{0x1F, {100.000000, 100.000000}},
			{0x21, {75.000000, 75.000000}},
			{0x22, {75.000000, 75.000000}},
			{0x23, {75.000000, 75.000000}},
			{0x26, {75.000000, 75.000000}},
			{0x29, {75.000000, 75.000000}},
			{0x2D, {75.000000, 75.000000}},
			{0x38, {4.000000, 4.250000}},
			{0x39, {2.000000, 2.000000}}
		};

		static Value offset = 1000.0;

		Value X, Y, Z, aX, aY, aZ;
		GetPos(player, X, Y, Z);
		GetAngle(player, aX, aY, aZ);

		AdvancePosition(X, Y, aZ, offset);

		ID ant = CreateActor(static_cast<Base>(antType), GetCell(player), X, Y, Z);

		aZ += 180.0;

		if (aZ > 360.0)
			aZ -= 360.0;
		else if (aZ < 0.00)
			aZ += 360.0;

		SetAngle(ant, aX, aY, aZ);

		for (const auto& value : values)
		{
			SetActorBaseValue(ant, static_cast<ActorValue>(value.first), value.second.first);
			SetActorValue(ant, static_cast<ActorValue>(value.first), value.second.second);
		}

		antsAI[ant] = AntState::ANT_IDLE;
		antsHealth[ant] = antHealth;
		CreateTimerEx(AntAI, static_cast<Interval>(100), ant);
		return False;
	}

	return True;
}

Void VAULTSCRIPT OnSpawn(ID object) noexcept
{
	Player player(object);

	if (player)
		player.UIMessage("Hello, " + player.GetBaseName() + "!");
}

Void VAULTSCRIPT OnActorValueChange(ID actor, ActorValue index, Value value) noexcept
{
	if (antsAI.find(actor) != antsAI.end() && index == health)
	{
		if (value < antsHealth[actor])
			PlayIdle(actor, antsReactions[std::rand() % antsReactions.size()]);
		antsHealth[actor] = value;
	}
}

void AntAttackRoutine(ID actor)
{
	Value X, Y, Z;
	GetPos(actor, X, Y, Z);

	for (auto& ant_ : antsAI)
	{
		Actor ant(ant_.first);

		if (ant && ant.GetCell() == GetCell(actor) && ant.IsNearPoint(X, Y, Z, antAggro) && (ant_.second == AntState::ANT_IDLE || ant_.second == AntState::ANT_ALERT))
		{
			ant_.second = AntState::ANT_AGGRO;
			antsTarget[ant_.first] = actor;
		}
	}
}

Void VAULTSCRIPT OnActorPunch(ID actor, State power) noexcept
{
	AntAttackRoutine(actor);
}

Void VAULTSCRIPT OnActorFireWeapon(ID actor, WEAP weapon) noexcept
{
	AntAttackRoutine(actor);
}
