#include "vaultscript.h"
#include <map>
#include <cstdio>
#include <cstring>
#include <cmath>
#include <ctime>
#include <cstdlib>

using namespace vaultmp;

enum class CarterState : unsigned int
{
	CARTER_IDLE = 0x01,
	CARTER_DETECT = CARTER_IDLE << 1,
};

static Base carterType = static_cast<Base>(NPC_::Carter);
static IDHash<CarterState> cartersAI;
static IDHash<ID> cartersTarget;

static Value carterStop = 50.0;

Void VAULTSCRIPT exec() noexcept
{
	std::srand(std::time(nullptr));
}

void AdvancePosition(Value& X, Value& Y, Value aZ, Value offset)
{
	X += (offset * std::sin((aZ / 360.0) * 2 * M_PI));
	Y += (offset * std::cos((aZ / 360.0) * 2 * M_PI));
}

Result RemoveObject(ID object_) noexcept
{
	DestroyObject(object_);
	cartersAI.erase(object_);
	cartersTarget.erase(object_);
	KillTimer();
	return static_cast<Result>(0);
}

Result CarterAI(ID) noexcept;

Result SleepTimer(ID object_) noexcept
{
	Item item(cartersTarget[object_]);

	if (item)
	{
		Actor actor(object_);

		if (actor)
		{
			actor.AddItem(item.GetBase(), item.GetItemCount(), item.GetItemCondition());
			item.DestroyObject();
		}
	}

	SetActorSneaking(object_, False);

	cartersTarget[object_] = static_cast<ID>(0);
	cartersAI[object_] = CarterState::CARTER_IDLE;

	KillTimer();
	CreateTimerEx(CarterAI, static_cast<Interval>(100), object_);
	return static_cast<Result>(0);
}

Result CarterAI(ID carter_) noexcept
{
	Actor carter(carter_);

	static Index idle = static_cast<Index>(0x00);

	if (carter)
	{
		if (!carter.GetActorDead())
		{
			if (cartersAI[carter_] == CarterState::CARTER_IDLE)
			{
				auto items = Item::GetList();

				for (const auto& item : items)
				{
					if (GetCell(item) == GetCell(carter_) && !GetItemContainer(item))
					{
						cartersTarget[carter_] = item;
						cartersAI[carter_] = CarterState::CARTER_DETECT;
						break;
					}
				}
			}
			else
			{
				Item item(cartersTarget[carter_]);

				if (item)
				{
					Value itemX, itemY, itemZ;
					Value carterX, carterY, carterZ, carterAX, carterAY, carterAZ;

					item.GetPos(itemX, itemY, itemZ);
					carter.GetPos(carterX, carterY, carterZ);
					carter.GetAngle(carterAX, carterAY, carterAZ);

					carterAZ = atan2(itemX - carterX, itemY - carterY) * 180 / M_PI;

					if (carterAZ < 0)
						carterAZ += 360.0;

					carter.SetAngle(carterAX, carterAY, carterAZ);

					if (!carter.IsNearPoint(itemX, itemY, carterZ, carterStop))
					{
						static Index fast_forward = static_cast<Index>(0x07);
						AdvancePosition(carterX, carterY, carterAZ, 22.0);
						carter.SetActorMovingAnimation(fast_forward);
						carter.SetPos(carterX, carterY, carterZ);
					}
					else
					{
						carter.SetActorMovingAnimation(idle);
						carter.SetActorSneaking(True);
						KillTimer();
						CreateTimerEx(SleepTimer, static_cast<Interval>(1500), carter_);
					}
				}
				else
				{
					cartersAI[carter_] = CarterState::CARTER_IDLE;
					carter.SetActorMovingAnimation(idle);
				}
			}
		}
		else
		{
			carter.SetActorMovingAnimation(idle);
			KillTimer();
			CreateTimerEx(RemoveObject, static_cast<Interval>(8000), carter_);
		}
	}
	else
		KillTimer();

	return static_cast<Result>(0);
}

State VAULTSCRIPT OnPlayerChat(ID player, RawString message) noexcept
{
	if (!std::strcmp(message, "/carter"))
	{
		static Value offset = 1000.0;

		Value X, Y, Z, aX, aY, aZ;
		GetPos(player, X, Y, Z);
		GetAngle(player, aX, aY, aZ);

		AdvancePosition(X, Y, aZ, offset);

		ID carter = CreateActor(static_cast<Base>(carterType), GetCell(player), X, Y, Z);
		RemoveAllItems(carter);

		aZ += 180.0;

		if (aZ > 360.0)
			aZ -= 360.0;
		else if (aZ < 0.00)
			aZ += 360.0;

		SetAngle(carter, aX, aY, aZ);

		cartersAI[carter] = CarterState::CARTER_IDLE;
		CreateTimerEx(SleepTimer, static_cast<Interval>(2000), carter);
		return False;
	}

	return True;
}

Void VAULTSCRIPT OnSpawn(ID object) noexcept
{
	Player player(object);

	if (player)
		player.UIMessage("Hello, " + player.GetBaseName() + "!");
}
